# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Device information view."""

import json
import logging
import threading
import urllib.request

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("Shumate", "1.0")
import contextlib

from gi.repository import Adw, Gdk, GLib, Gtk, Pango, Shumate
from meshy.models import DeviceInfo, snr_to_icon

log = logging.getLogger(__name__)

_RELEASES_URL = "https://flasher.meshcore.io/releases"
_FLASHER_URL = "https://flasher.meshcore.io"


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/device-view.ui")
class DeviceView(Gtk.Box):
    __gtype_name__ = "MeshyDeviceView"

    _scroll = Gtk.Template.Child("scroll")
    _status_group = Gtk.Template.Child("status_group")
    _status_row = Gtk.Template.Child("status_row")
    _status_icon = Gtk.Template.Child("status_icon")
    _info_group = Gtk.Template.Child("info_group")
    _name_row = Gtk.Template.Child("name_row")
    _board_row = Gtk.Template.Child("board_row")
    _fw_row = Gtk.Template.Child("fw_row")
    _fw_update_btn = Gtk.Template.Child("fw_update_btn")
    _key_row = Gtk.Template.Child("key_row")
    _copy_key_btn = Gtk.Template.Child("copy_key_btn")
    _telemetry_row = Gtk.Template.Child("telemetry_row")
    _telemetry_btn = Gtk.Template.Child("telemetry_btn")
    _location_row = Gtk.Template.Child("location_row")
    _location_map_btn = Gtk.Template.Child("location_map_btn")
    _batt_group = Gtk.Template.Child("batt_group")
    _batt_row = Gtk.Template.Child("batt_row")
    _batt_bar = Gtk.Template.Child("batt_bar")
    _storage_row = Gtk.Template.Child("storage_row")
    _storage_bar = Gtk.Template.Child("storage_bar")
    _radio_group = Gtk.Template.Child("radio_group")
    _freq_row = Gtk.Template.Child("freq_row")
    _bw_row = Gtk.Template.Child("bw_row")
    _sf_row = Gtk.Template.Child("sf_row")
    _cr_row = Gtk.Template.Child("cr_row")
    _tx_row = Gtk.Template.Child("tx_row")
    _path_hash_row = Gtk.Template.Child("path_hash_row")
    _stats_group = Gtk.Template.Child("stats_group")
    _uptime_row = Gtk.Template.Child("uptime_row")
    _queue_row = Gtk.Template.Child("queue_row")
    _noise_row = Gtk.Template.Child("noise_row")
    _rssi_row = Gtk.Template.Child("rssi_row")
    _snr_row = Gtk.Template.Child("snr_row")
    _airtime_row = Gtk.Template.Child("airtime_row")
    _pkt_row = Gtk.Template.Child("pkt_row")
    _pkt_flood_row = Gtk.Template.Child("pkt_flood_row")
    _pkt_direct_row = Gtk.Template.Child("pkt_direct_row")
    _refresh_stats_btn = Gtk.Template.Child("refresh_stats_btn")

    def __init__(self, window, **kwargs):
        super().__init__(**kwargs)
        self._window = window

        self._location_lat = None
        self._location_lon = None
        self._location_pending_lat = None
        self._location_pending_lon = None
        self._gps_enabled = None  # None = unknown, True/False from custom_vars

        self._section_widgets = [
            ("Connection", self._status_group),
            ("Device Information", self._info_group),
            ("Battery &amp; Storage", self._batt_group),
            ("Radio Configuration", self._radio_group),
            ("Statistics", self._stats_group),
        ]

        self._latest_fw_version = None

        # Connect signals that reference self._window
        self._copy_key_btn.connect("clicked", self._on_copy_public_key)
        self._telemetry_btn.connect("clicked", self._on_request_telemetry)
        self._location_map_btn.connect("clicked", self._on_show_location_map)
        self._fw_update_btn.connect("clicked", self._on_fw_update_clicked)
        self._refresh_stats_btn.connect("clicked", lambda *_: self._window.refresh_stats())

    def _build_actions_group(self) -> Adw.PreferencesGroup:
        """Build the Actions preference group from UI template."""
        builder = Gtk.Builder.new_from_resource(
            "/page/codeberg/sesivany/Meshy/ui/device-actions.ui"
        )
        actions_group = builder.get_object("actions_group")

        builder.get_object("zero_hop_btn").connect(
            "clicked", lambda *_: self._window.send_advert(flood=False)
        )
        builder.get_object("flood_btn").connect(
            "clicked", lambda *_: self._window.send_advert(flood=True)
        )
        builder.get_object("clipboard_btn").connect("clicked", self._on_advert_to_clipboard)
        builder.get_object("qr_btn").connect("clicked", self._on_show_qr_code)
        builder.get_object("trace_btn").connect("clicked", self._on_trace_path)
        builder.get_object("discover_btn").connect("clicked", self._on_discover_nearby)
        builder.get_object("rx_log_btn").connect("clicked", self._on_rx_log)
        builder.get_object("reboot_btn").connect("clicked", self._on_reboot)

        return actions_group

    def build_sidebar_actions(self) -> Gtk.Widget:
        """Build a fresh Actions group for placement in the sidebar."""
        return self._build_actions_group()

    @property
    def sections(self) -> list[str]:
        return [name for name, _w in self._section_widgets]

    def scroll_to_section(self, section_name: str):
        for name, widget in self._section_widgets:
            if name == section_name:
                widget.grab_focus()
                # Scroll the widget into view
                GLib.idle_add(self._do_scroll_to, widget)
                break

    def _do_scroll_to(self, widget):
        # Get widget position relative to the scrolled window
        success, x, y = widget.translate_coordinates(self._scroll.get_child(), 0, 0)
        if success:
            adj = self._scroll.get_vadjustment()
            adj.set_value(y)
        return False

    def _on_copy_public_key(self, *args):
        key = self._key_row.get_subtitle()
        if key and key != "--":
            clipboard = self._window.get_clipboard()
            clipboard.set(key)
            self._window.show_toast(_("Public key copied"))

    def _on_advert_to_clipboard(self, *args):
        """Export self contact and copy to clipboard."""
        self._window.export_self_to_clipboard()

    def _on_show_qr_code(self, *args):
        """Show a QR code dialog with structured meshcore:// URL."""
        info = self._window.device_info
        if not info.public_key:
            self._window.show_toast(_("Device public key not available"))
            return
        self._show_qr_dialog(None)

    def _show_qr_dialog(self, _unused=None):
        """Display a QR code dialog for sharing this device as a contact."""
        import io

        try:
            import segno
        except ImportError:
            self._window.show_toast(_("QR code library not available"))
            return

        dialog = Adw.Dialog()
        dialog.set_title(_("Your Contact QR Code"))
        dialog.set_content_width(380)
        dialog.set_content_height(440)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=12,
            margin_start=24,
            margin_end=24,
            margin_top=12,
            margin_bottom=24,
        )

        # Device name
        info = self._window.device_info
        name = info.name or "My Device"
        name_label = Gtk.Label(label=name)
        name_label.add_css_class("title-2")
        content.append(name_label)

        # Build structured URL (accepted by official client)
        from urllib.parse import quote

        structured_uri = (
            f"meshcore://contact/add?name={quote(name)}" f"&public_key={info.public_key}&type=1"
        )

        qr = segno.make(structured_uri)
        png_buffer = io.BytesIO()
        qr.save(png_buffer, kind="png", scale=8, border=2)
        texture = Gdk.Texture.new_from_bytes(GLib.Bytes.new(png_buffer.getvalue()))

        picture = Gtk.Picture.new_for_paintable(texture)
        picture.set_content_fit(Gtk.ContentFit.CONTAIN)
        picture.set_size_request(280, 280)
        content.append(picture)

        # Instructions
        hint = Gtk.Label(label=_("Scan this QR code to add this contact"))
        hint.add_css_class("dim-label")
        content.append(hint)

        toolbar_view.set_content(content)
        dialog.set_child(toolbar_view)
        dialog.present(self._window)

    def _on_factory_reset(self, *args):
        dialog = Adw.AlertDialog(
            heading=_("Factory Reset?"),
            body=_(
                "This will permanently erase ALL data on the companion "
                "device, including contacts, messages, channels, settings, "
                "and the device identity (keys). "
                "This action cannot be undone."
            ),
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("reset", _("Factory Reset"))
        dialog.set_response_appearance("reset", Adw.ResponseAppearance.DESTRUCTIVE)

        def _on_response(d, response):
            if response == "reset":
                self._window.factory_reset_device()

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _on_reboot(self, *args):
        dialog = Adw.AlertDialog(
            heading=_("Reboot Device?"),
            body=_("The device will disconnect and restart."),
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("reboot", _("Reboot"))
        dialog.set_response_appearance("reboot", Adw.ResponseAppearance.DESTRUCTIVE)

        def _on_response(d, response):
            if response == "reboot":
                self._window.reboot_device()

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def update_device_info(self, info: DeviceInfo):
        self._name_row.set_subtitle(info.name or "--")
        self._board_row.set_subtitle(info.model or "--")
        self._fw_row.set_subtitle(
            f"{info.ver} ({info.fw_build})" if info.ver and info.fw_build else info.ver or "--"
        )
        if info.ver:
            self._check_firmware_update(info.ver)
        if info.public_key:
            self._key_row.set_subtitle(info.public_key)

        # Battery
        if info.battery_mv > 0:
            pct = info.battery_percent
            self._batt_row.set_subtitle(f"{pct}% ({info.battery_mv} mV)")
            self._batt_bar.set_value(pct)

        # Storage
        if info.storage_total_kb > 0:
            spct = info.storage_percent
            self._storage_row.set_subtitle(
                f"{spct}% ({info.storage_used_kb} / {info.storage_total_kb} KB)"
            )
            self._storage_bar.set_value(spct)

        # Radio - radio_freq is in MHz, radio_bw is in kHz
        if info.radio_freq:
            self._freq_row.set_subtitle(f"{info.radio_freq:.3f} MHz")
        if info.radio_bw:
            self._bw_row.set_subtitle(f"{info.radio_bw:.1f} kHz")
        if info.radio_sf:
            self._sf_row.set_subtitle(f"SF{info.radio_sf}")
        if info.radio_cr:
            self._cr_row.set_subtitle(f"4/{info.radio_cr}")
        if info.tx_power:
            self._tx_row.set_subtitle(f"{info.tx_power} dBm")
        hash_labels = {
            0: _("1-byte (max 64 hops)"),
            1: _("2-byte (max 32 hops)"),
            2: _("3-byte (max 21 hops)"),
        }
        self._path_hash_row.set_subtitle(hash_labels.get(info.path_hash_mode, "--"))

    def update_location(self, lat: float, lon: float):
        """Update the location row from the device's advertised position."""
        if self._gps_enabled is None:
            # Custom vars not yet received — store for later
            self._location_pending_lat = lat
            self._location_pending_lon = lon
            return
        self._apply_location(lat, lon)

    def update_gps_enabled(self, enabled: bool):
        """Update GPS enabled state from custom_vars."""
        self._gps_enabled = enabled
        # Apply pending location now that we know GPS state
        if self._location_pending_lat is not None:
            self._apply_location(self._location_pending_lat, self._location_pending_lon)
            self._location_pending_lat = None
            self._location_pending_lon = None
        else:
            self._update_location_title()

    def _apply_location(self, lat: float, lon: float):
        has_pos = abs(lat) > 1e-6 or abs(lon) > 1e-6
        if has_pos:
            self._location_lat = lat
            self._location_lon = lon
            self._location_row.set_subtitle(self._format_location(lat, lon))
        else:
            self._location_lat = None
            self._location_lon = None
            self._location_row.set_subtitle(_("Not set"))
        self._location_map_btn.set_visible(has_pos)
        self._update_location_title()

    def _update_location_title(self):
        has_pos = self._location_lat is not None
        if self._gps_enabled is None:
            # No GPS hardware on companion
            self._location_row.set_title(_("Location (manually set)") if has_pos else _("Location"))
        elif self._gps_enabled:
            self._location_row.set_title(
                _("Location (GPS)") if has_pos else _("Location (GPS on, no fix)")
            )
        else:
            # GPS off but has position — set manually or via system location
            self._location_row.set_title(
                _("Location (manually set)") if has_pos else _("Location (GPS off)")
            )

    @staticmethod
    def _format_location(lat: float, lon: float) -> str:
        lat_dir = "N" if lat >= 0 else "S"
        lon_dir = "E" if lon >= 0 else "W"
        return f"{abs(lat):.6f}° {lat_dir}, {abs(lon):.6f}° {lon_dir}"

    def _on_request_telemetry(self, btn):
        """Request telemetry from the connected companion device."""
        from meshy.protocol import (
            LPP_ALTITUDE,
            LPP_GPS,
            LPP_TEMPERATURE,
            LPP_VOLTAGE,
            parse_cayenne_lpp,
        )

        pub_key_hex = self._window.device_info.public_key
        if not pub_key_hex:
            self._telemetry_row.set_subtitle(_("Not connected"))
            return

        btn.set_sensitive(False)
        self._telemetry_row.set_subtitle(_("Requesting…"))

        timeout_id = [None]

        def on_timeout():
            btn.set_sensitive(True)
            self._telemetry_row.set_subtitle(_("No response (timed out)"))
            timeout_id[0] = None
            return False

        def on_response(lpp_data):
            btn.set_sensitive(True)
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])
                timeout_id[0] = None
            if not lpp_data:
                self._telemetry_row.set_subtitle(_("No telemetry data"))
                return
            entries = parse_cayenne_lpp(lpp_data)
            if not entries:
                self._telemetry_row.set_subtitle(_("No telemetry data"))
                return

            # Build summary for the subtitle
            parts = []
            for e in entries:
                if e["type"] == LPP_TEMPERATURE:
                    parts.append(f"{e['value']:.1f} °C")
                elif e["type"] == LPP_ALTITUDE:
                    parts.append(f"{e['value']} m alt")
                elif e["type"] == LPP_VOLTAGE:
                    parts.append(f"{e['value']:.2f} V")
                elif e["type"] == LPP_GPS:
                    val = e["value"]
                    parts.append(f"{val['alt']:.0f} m alt")
            self._telemetry_row.set_subtitle(", ".join(parts) if parts else _("Telemetry received"))

        timeout_id[0] = GLib.timeout_add(15000, on_timeout)
        self._window.get_self_telemetry(on_response)

    def _on_show_location_map(self, btn):
        if self._location_lat is None or self._location_lon is None:
            return
        lat, lon = self._location_lat, self._location_lon
        name = self._window.device_info.name or "My Device"

        dialog = Adw.Dialog()
        dialog.set_title(_("Device Location"))
        dialog.set_content_width(400)
        dialog.set_content_height(450)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=8,
            margin_start=12,
            margin_end=12,
            margin_bottom=12,
        )

        # Coordinates label
        coords = Gtk.Label(label=self._format_location(lat, lon))
        coords.add_css_class("caption")
        content.append(coords)

        # Map
        from meshy.views import create_shumate_map, deregister_map_widget
        from meshy.views.map_view import _SELF_COLOR, MapView

        map_widget, marker_layer, viewport = create_shumate_map(zoom_level=14, lat=lat, lon=lon)
        map_widget.set_size_request(-1, 350)

        marker = MapView.create_marker(*_SELF_COLOR, "avatar-default-symbolic", name, lat, lon)
        marker_layer.add_marker(marker)

        map_frame = Gtk.Frame()
        map_frame.set_child(map_widget)
        content.append(map_frame)

        toolbar_view.set_content(content)
        dialog.set_child(toolbar_view)
        dialog.connect("closed", lambda _d: deregister_map_widget(map_widget))
        dialog.present(self._window)

    def update_stats(self, stats: dict):
        """Update statistics display from CMD_GET_STATS responses."""
        from meshy.protocol import STATS_TYPE_CORE, STATS_TYPE_PACKETS, STATS_TYPE_RADIO

        core = stats.get(STATS_TYPE_CORE, {})
        if core:
            secs = core.get("uptime_secs", 0)
            days, rem = divmod(secs, 86400)
            hours, rem = divmod(rem, 3600)
            mins, _secs = divmod(rem, 60)
            parts = []
            if days:
                parts.append(_("{days}d").format(days=days))
            if hours:
                parts.append(_("{hours}h").format(hours=hours))
            parts.append(_("{mins}m").format(mins=mins))
            self._uptime_row.set_subtitle(" ".join(parts))
            self._queue_row.set_subtitle(_("{} messages").format(core.get("queue_len", 0)))

        radio = stats.get(STATS_TYPE_RADIO, {})
        if radio:
            self._noise_row.set_subtitle(f'{radio.get("noise_floor", 0)} dBm')
            self._rssi_row.set_subtitle(f'{radio.get("last_rssi", 0)} dBm')
            self._snr_row.set_subtitle(f'{radio.get("last_snr", 0):.1f} dB')
            tx_s = radio.get("tx_air_secs", 0)
            rx_s = radio.get("rx_air_secs", 0)
            self._airtime_row.set_subtitle(f"TX: {tx_s}s / RX: {rx_s}s")

        pkts = stats.get(STATS_TYPE_PACKETS, {})
        if pkts:
            recv = pkts.get("recv", 0)
            sent = pkts.get("sent", 0)
            errs = pkts.get("recv_errors")
            sub = f"RX: {recv} / TX: {sent}"
            if errs is not None:
                sub += " / " + _("{count} errors").format(count=errs)
            self._pkt_row.set_subtitle(sub)
            self._pkt_flood_row.set_subtitle(
                f'TX: {pkts.get("flood_tx", 0)} / RX: {pkts.get("flood_rx", 0)}'
            )
            self._pkt_direct_row.set_subtitle(
                f'TX: {pkts.get("direct_tx", 0)} / RX: {pkts.get("direct_rx", 0)}'
            )

    def update_connection_status(self, connected: bool, transport: str = ""):
        """Update the connection status icon and text."""
        if connected:
            transport_labels = {"ble": "Bluetooth", "usb": "USB", "tcp": "TCP"}
            label = transport_labels.get(transport, "")
            if label:
                self._status_row.set_subtitle("{} \u00b7 {}".format(label, _("Connected")))
            else:
                self._status_row.set_subtitle(_("Connected"))
            self._status_icon.set_from_icon_name("charge-symbolic")
        else:
            self._status_row.set_subtitle(_("Disconnected"))
            self._status_icon.set_from_icon_name("network-offline-symbolic")

    def _on_trace_path(self, *args):
        """Open the Trace Path dialog."""
        from meshy.models import ContactType

        # Trace protocol only supports 1B and 2B hashes (power-of-2 encoding)
        if self._window.device_info.path_hash_mode > 1:
            self._window.show_toast(_("Trace Path is not supported with 3-byte path hashes"))
            return

        dialog = Adw.Dialog()
        dialog.set_title(_("Trace Path"))
        dialog.set_content_width(460)
        dialog.set_content_height(500)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=8,
            margin_start=8,
            margin_end=8,
            margin_bottom=8,
        )

        # Path hash size: mode 0→1 byte, mode 1→2 bytes, mode 2→3 bytes
        hash_width = self._window.device_info.path_hash_mode + 1
        examples = {1: _("e.g. aa,bb,cc"), 2: _("e.g. aabb,ccdd"), 3: _("e.g. aabbcc,ddeeff")}

        # Own node hash for bookending the trace
        self_pub_key = (
            bytes.fromhex(self._window.device_info.public_key)
            if self._window.device_info.public_key
            else b""
        )
        self_hash = self_pub_key[:hash_width] if len(self_pub_key) >= hash_width else b""
        self_name = self._window.device_info.name or "This device"

        # Add from Contacts button (above path entry)
        add_contact_btn = Gtk.MenuButton(label=_("Add from Contacts"))
        add_contact_btn.add_css_class("flat")
        add_contact_btn.set_halign(Gtk.Align.START)

        popover = Gtk.Popover()
        popover.set_size_request(300, -1)
        pop_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=4)
        pop_box.set_margin_start(8)
        pop_box.set_margin_end(8)
        pop_box.set_margin_top(8)
        pop_box.set_margin_bottom(8)

        search_entry = Gtk.SearchEntry(placeholder_text="Search repeaters...")
        pop_box.append(search_entry)

        repeater_list = Gtk.ListBox()
        repeater_list.set_selection_mode(Gtk.SelectionMode.NONE)
        repeater_list.add_css_class("boxed-list")

        repeaters = [
            c for c in self._window.contacts if c.type in (ContactType.REPEATER, ContactType.ROOM)
        ]
        repeater_rows = []
        for contact in repeaters:
            hop_hex = contact.public_key[:hash_width].hex()
            row = Adw.ActionRow(title=GLib.markup_escape_text(contact.name), subtitle=hop_hex)
            row.set_activatable(True)
            row._hop_hex = hop_hex
            row._search_text = contact.name.lower()
            repeater_list.append(row)
            repeater_rows.append(row)

        if not repeaters:
            empty = Gtk.Label(label=_("No repeaters or rooms in contacts"))
            empty.add_css_class("dim-label")
            empty.set_margin_top(8)
            empty.set_margin_bottom(8)
            pop_box.append(empty)

        # Pick on Map button
        pick_map_btn = Gtk.Button(label=_("Pick on Map"))
        pick_map_btn.add_css_class("flat")
        has_located = any(c.has_location for c in repeaters)
        pick_map_btn.set_sensitive(has_located)
        if not has_located:
            pick_map_btn.set_tooltip_text(_("No repeaters with known location"))

        # Path input
        path_entry = Adw.EntryRow(title=_("Path"))
        path_entry.set_text("")

        btn_box = Gtk.Box(spacing=8)
        btn_box.append(add_contact_btn)
        pick_map_btn.set_hexpand(True)
        pick_map_btn.set_halign(Gtk.Align.END)
        btn_box.append(pick_map_btn)
        content.append(btn_box)
        content.append(path_entry)

        placeholder_label = Gtk.Label(label=examples.get(hash_width, examples[1]))
        placeholder_label.add_css_class("dim-label")
        placeholder_label.add_css_class("caption")
        placeholder_label.set_halign(Gtk.Align.START)
        placeholder_label.set_margin_start(12)
        content.append(placeholder_label)

        def on_search_changed(entry):
            query = entry.get_text().strip().lower()
            for row in repeater_rows:
                row.set_visible(query in row._search_text or query in row._hop_hex)

        search_entry.connect("search-changed", on_search_changed)

        def on_row_activated(lb, row):
            current = path_entry.get_text().strip()
            if current:
                path_entry.set_text(f"{current},{row._hop_hex}")
            else:
                path_entry.set_text(row._hop_hex)
            popover.popdown()

        repeater_list.connect("row-activated", on_row_activated)

        repeater_scroll = Gtk.ScrolledWindow()
        repeater_scroll.set_child(repeater_list)
        repeater_scroll.set_min_content_height(120)
        repeater_scroll.set_max_content_height(300)
        pop_box.append(repeater_scroll)

        popover.set_child(pop_box)
        add_contact_btn.set_popover(popover)

        def on_pick_map(*args):
            from meshy.views.map_view import open_trace_picker_map

            def _on_map_done(hex_string):
                current = path_entry.get_text().strip()
                if current:
                    path_entry.set_text(f"{current},{hex_string}")
                else:
                    path_entry.set_text(hex_string)

            open_trace_picker_map(self._window, self._window.contacts, hash_width, _on_map_done)

        pick_map_btn.connect("clicked", on_pick_map)

        # Run button
        run_btn = Gtk.Button(label=_("Run Trace"))
        run_btn.add_css_class("suggested-action")
        run_btn.set_halign(Gtk.Align.CENTER)
        run_btn.set_margin_top(4)
        content.append(run_btn)

        # Separator
        content.append(Gtk.Separator())

        # Status
        status_box = Gtk.Box(spacing=8, halign=Gtk.Align.CENTER)
        spinner = Gtk.Spinner()
        status_box.append(spinner)
        status_label = Gtk.Label(label=_("Enter path hashes and press Run Trace"))
        status_label.add_css_class("dim-label")
        status_box.append(status_label)
        status_map_btn = Gtk.Button(icon_name="map-symbolic")
        status_map_btn.add_css_class("flat")
        status_map_btn.set_tooltip_text("Show on Map")
        status_map_btn.set_visible(False)
        status_box.append(status_map_btn)
        content.append(status_box)

        # Results list
        results_scroll = Gtk.ScrolledWindow(vexpand=True)
        results_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        results_scroll.set_child(results_box)
        content.append(results_scroll)

        toolbar_view.set_content(content)
        dialog.set_child(toolbar_view)

        timeout_id = [None]
        trace_start = [0.0]

        def clear_results():
            child = results_box.get_first_child()
            while child:
                next_c = child.get_next_sibling()
                results_box.remove(child)
                child = next_c

        def add_node_row(label):
            """Add a node row to results."""
            row = Gtk.Label(label=label, xalign=0)
            row.set_margin_start(16)
            row.set_margin_top(6)
            row.set_margin_bottom(2)
            results_box.append(row)

        def add_snr_row(snr):
            """Add an SNR indicator row between nodes."""
            snr_text = f"\u2502  {snr:+.1f} dB"
            label = Gtk.Label(label=snr_text, xalign=0)
            label.set_margin_start(20)
            label.set_margin_top(2)
            label.set_margin_bottom(2)
            label.add_css_class("caption")
            if snr >= 0:
                label.add_css_class("success")
            elif snr >= -5:
                label.add_css_class("dim-label")
            else:
                label.add_css_class("error")
            results_box.append(label)

        def show_results(result):
            """Display trace results in the results box."""
            spinner.set_spinning(False)
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])
                timeout_id[0] = None

            hops = result["path_hops"]  # list of bytes objects (per-hop hashes)
            snr_values = result["snr_values"]
            clear_results()

            import time

            duration = time.monotonic() - trace_start[0]
            status_label.set_label(
                _("Trace complete: {} hops, {:.1f}s").format(len(hops), duration)
            )

            hop_hexes = [h.hex() for h in hops]
            hop_contacts = self._window.resolve_hop_prefixes(hop_hexes)

            # Start with companion node
            add_node_row(f"\u25cf {self_name} ({self_hash.hex()})")

            for i, hop_bytes in enumerate(hops):
                # SNR between previous node and this one
                if i < len(snr_values):
                    add_snr_row(snr_values[i])

                hop_hex = hop_bytes.hex()
                contact = hop_contacts[i]
                if contact:
                    add_node_row(f"\u25cf {contact.name} ({hop_hex})")
                else:
                    add_node_row(f"\u25cb {hop_hex}")

            # End with companion node (return path)
            if len(snr_values) > len(hops):
                add_snr_row(snr_values[len(hops)])
            add_node_row(f"\u25cf {self_name} ({self_hash.hex()})")

            # Show on Map button in status bar if any hop has a known location
            has_any_location = any(c and c.has_location for c in hop_contacts)
            if has_any_location:
                status_map_btn.set_visible(True)
                # Reconnect handler for new results
                with contextlib.suppress(AttributeError, TypeError):
                    status_map_btn.disconnect_by_func(status_map_btn._handler)

                def _on_map_clicked(
                    *_, hc=hop_contacts, snr=snr_values, hw=hash_width, hp=hop_hexes
                ):
                    self._show_trace_on_map(hc, snr, hw, hp)

                status_map_btn._handler = _on_map_clicked
                status_map_btn.connect("clicked", _on_map_clicked)
            else:
                status_map_btn.set_visible(False)

            run_btn.set_sensitive(True)

        def on_timeout():
            spinner.set_spinning(False)
            status_label.set_label(_("Trace timed out"))
            self._window.stop_trace_path()
            run_btn.set_sensitive(True)
            timeout_id[0] = None
            return False

        def on_run(*args):
            text = path_entry.get_text().strip()
            if not text:
                return
            parts = [p.strip() for p in text.split(",") if p.strip()]
            try:
                path_bytes = b""
                for p in parts:
                    path_bytes += bytes.fromhex(p)
            except ValueError:
                status_label.set_label(_("Invalid hex in path"))
                return

            trace_bytes = path_bytes

            clear_results()
            spinner.set_spinning(True)
            status_label.set_label(_("Tracing..."))
            status_map_btn.set_visible(False)
            run_btn.set_sensitive(False)
            import time

            trace_start[0] = time.monotonic()

            def on_device_timeout(timeout_ms):
                """Device told us how long to wait."""
                if timeout_id[0]:
                    GLib.source_remove(timeout_id[0])
                timeout_id[0] = GLib.timeout_add(timeout_ms, on_timeout)

            # Start with a fallback timeout; device may override with its own
            timeout_id[0] = GLib.timeout_add(30000, on_timeout)
            # Flag encodes hash size: mode 0→flag 0 (1B), mode 1→flag 1 (2B)
            trace_flags = self._window.device_info.path_hash_mode
            self._window.send_trace_path(
                trace_bytes, show_results, on_device_timeout, flags=trace_flags
            )

        run_btn.connect("clicked", on_run)

        def on_dialog_closed(d):
            self._window.stop_trace_path()
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])

        dialog.connect("closed", on_dialog_closed)
        dialog.present(self._window)

    def _show_trace_on_map(self, hop_contacts, snr_values, hash_width, hop_prefixes=None):
        """Show trace path results in a map dialog."""
        from meshy.views.map_view import (
            _MARKER_COLORS,
            _MARKER_ICONS,
            _SELF_COLOR,
            _UNKNOWN_COLOR,
            MapView,
            companion_too_far,
        )

        device = self._window.device_info
        self_lat = getattr(device, "latitude", None)
        self_lon = getattr(device, "longitude", None)
        has_self = self_lat and self_lon and (abs(self_lat) > 1e-6 or abs(self_lon) > 1e-6)

        # Build full hop list: (lat, lon, contact, prefix, snr)
        raw_hops = []
        for i, contact in enumerate(hop_contacts):
            snr = snr_values[i] if i < len(snr_values) else None
            prefix = hop_prefixes[i] if hop_prefixes and i < len(hop_prefixes) else None
            if contact and contact.has_location:
                raw_hops.append((contact.latitude, contact.longitude, contact, prefix, snr))
            else:
                raw_hops.append((None, None, contact, prefix, snr))

        if has_self and not companion_too_far(self_lat, self_lon, raw_hops):
            raw_hops.insert(0, (self_lat, self_lon, None, None, None))

        hops = MapView.interpolate_unknown_hops(raw_hops)

        if not any(h[0] is not None for h in hops):
            return

        dialog, map_widget, marker_layer, viewport = MapView.open_map_dialog(
            self._window, title=_("Trace Path")
        )

        map_widget.get_map().remove_layer(marker_layer)
        path_layer = MapView.create_dashed_path_layer(viewport)
        map_widget.get_map().add_layer(path_layer)
        map_widget.get_map().add_layer(marker_layer)

        lats, lons = [], []
        hop_num = 0
        for lat, lon, contact, prefix, snr in hops:
            if lat is None:
                continue
            node = Shumate.Marker()
            node.set_location(lat, lon)
            path_layer.add_node(node)

            if contact is None and prefix is None:
                r, g, b = _SELF_COLOR
                icon_name = "network-wireless-symbolic"
                name = device.name or _("My Device")
                type_label = None
                badge = ""
            elif contact is None or not contact.has_location:
                r, g, b = _UNKNOWN_COLOR
                icon_name = "network-server-symbolic"
                name = prefix or _("Unknown")
                type_label = _("Unknown repeater")
                hop_num += 1
                badge = str(hop_num)
            else:
                r, g, b = _MARKER_COLORS.get(contact.type, (0.5, 0.5, 0.5))
                icon_name = _MARKER_ICONS.get(contact.type, "avatar-default-symbolic")
                name = contact.name
                type_label = contact.type_label
                hop_num += 1
                badge = str(hop_num)

            is_interpolated = (contact is None or not contact.has_location) and prefix is not None
            detail_cb = (lambda c=contact: self._window.show_contact_detail(c)) if contact else None
            widget = MapView._build_marker_widget(r, g, b, icon_name, "", badge_text=badge)
            widget.set_tooltip_text(name)
            click = Gtk.GestureClick()
            click.connect(
                "pressed",
                lambda g,
                n,
                x,
                y,
                w=widget,
                nm=name,
                la=lat,
                lo=lon,
                s=snr,
                t=type_label,
                ip=is_interpolated,
                od=detail_cb: MapView.show_marker_popover(
                    w, nm, la, lo, s, t, interpolated=ip, on_detail=od
                ),
            )
            widget.add_controller(click)

            hop_marker = Shumate.Marker()
            hop_marker.set_child(widget)
            hop_marker.set_location(lat, lon)
            marker_layer.add_marker(hop_marker)

            lats.append(lat)
            lons.append(lon)

        # SNR labels at midpoints
        for i in range(1, len(hops)):
            if hops[i][0] is None or hops[i - 1][0] is None:
                continue
            snr = hops[i][4]
            if snr is None:
                continue
            mid_lat = (hops[i - 1][0] + hops[i][0]) / 2
            mid_lon = (hops[i - 1][1] + hops[i][1]) / 2
            snr_widget = MapView.build_snr_widget(snr)
            snr_marker = Shumate.Marker()
            snr_marker.set_child(snr_widget)
            snr_marker.set_location(mid_lat, mid_lon)
            marker_layer.add_marker(snr_marker)

        MapView.fit_viewport(viewport, lats, lons)
        dialog.present(self._window)

    def _on_discover_nearby(self, *args):
        """Send a NODE_DISCOVER_REQ and show responding nodes live."""
        from meshy.models import ContactType

        dialog = Adw.Dialog()
        dialog.set_title(_("Discover Nearby Nodes"))
        dialog.set_content_width(420)
        dialog.set_content_height(480)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=8,
            margin_start=8,
            margin_end=8,
            margin_bottom=8,
        )

        # Status with spinner
        status_box = Gtk.Box(spacing=8, halign=Gtk.Align.CENTER)
        spinner = Gtk.Spinner(spinning=True)
        status_box.append(spinner)
        status_label = Gtk.Label(label=_("Scanning... {}s remaining").format(30))
        status_label.add_css_class("dim-label")
        status_box.append(status_label)
        content.append(status_box)

        # List of responding nodes
        scroll = Gtk.ScrolledWindow(vexpand=True)
        list_wrapper = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            margin_start=8,
            margin_end=8,
            margin_top=8,
            margin_bottom=8,
        )
        list_box = Gtk.ListBox()
        list_box.set_selection_mode(Gtk.SelectionMode.NONE)
        list_box.add_css_class("boxed-list")

        placeholder = Adw.StatusPage(
            icon_name="edit-find-symbolic",
            title=_("Listening..."),
            description=_("Waiting for nearby nodes to respond"),
        )
        list_box.set_placeholder(placeholder)

        list_wrapper.append(list_box)
        scroll.set_child(list_wrapper)
        content.append(scroll)

        toolbar_view.set_content(content)
        dialog.set_child(toolbar_view)

        seen = set()
        seconds_left = [30]
        timer_id = [None]

        type_icons = {
            ContactType.CHAT: "avatar-default-symbolic",
            ContactType.REPEATER: "network-server-symbolic",
            ContactType.ROOM: "user-home-symbolic",
            ContactType.SENSOR: "weather-clear-symbolic",
        }
        type_labels = {
            ContactType.CHAT: _("Chat"),
            ContactType.REPEATER: _("Repeater"),
            ContactType.ROOM: _("Room Server"),
            ContactType.SENSOR: _("Sensor"),
        }

        # Track rows by key prefix so adverts can update discover rows
        rows_by_prefix = {}  # key_prefix → row

        # Companion location for distance calculation
        dev = self._window.device_info
        self_lat = dev.latitude if dev else None
        self_lon = dev.longitude if dev else None
        has_self_loc = (
            self_lat is not None
            and self_lon is not None
            and (abs(self_lat) > 1e-6 or abs(self_lon) > 1e-6)
        )
        log.debug("Discover: companion loc=%s/%s has_self_loc=%s", self_lat, self_lon, has_self_loc)

        def _distance_str(contact):
            if not has_self_loc or not contact or not contact.has_location:
                return ""
            from meshy.utils import haversine_distance

            dist = haversine_distance(self_lat, self_lon, contact.latitude, contact.longitude)
            if dist < 1.0:
                return f"{dist * 1000:.0f} m"
            return f"{dist:.1f} km"

        def _add_node_button(row):
            """Create an Add to Contacts button for unknown nodes."""
            btn = Gtk.Button(icon_name="list-add-symbolic", valign=Gtk.Align.CENTER)
            btn.set_tooltip_text(_("Add to Contacts"))

            def on_add(_btn):
                info = row._node_info
                self._window.add_contact(info["pub_key"], info["type"], info["name"])
                _btn.set_sensitive(False)
                _btn.set_icon_name("object-select-symbolic")
                self._window.show_toast(_("Added {}").format(info["name"]))

            btn.connect("clicked", on_add)
            row.add_suffix(btn)

        def on_discover_response(result):
            """Handle a NODE_DISCOVER_RESP."""
            pub_key_hex = result.get("pub_key_hex", "")
            if not pub_key_hex:
                return
            if pub_key_hex in seen:
                return
            seen.add(pub_key_hex)

            snr = result.get("snr", 0)
            node_type = result.get("node_type", 1)
            snr_str = f"SNR: {snr:.1f} dB"

            contact = self._window.find_contact_by_key_prefix(pub_key_hex)
            is_known = contact is not None and self._window.is_contact_known(contact.public_key_hex)
            if contact:
                name = contact.name
                label = f"{contact.type_label} \u00b7 {snr_str}"
                icon = type_icons.get(contact.type, "avatar-default-symbolic")
                log.debug(
                    "Discover resp: %s found has_loc=%s", pub_key_hex[:16], contact.has_location
                )
            else:
                name = f"0x{pub_key_hex[:16].upper()}"
                label = "{} \u00b7 {}".format(type_labels.get(node_type, _("Node")), snr_str)
                icon = type_icons.get(node_type, "network-server-symbolic")
                log.debug("Discover resp: %s not found", pub_key_hex[:16])

            dist = _distance_str(contact)
            if dist:
                label += f" \u00b7 {dist}"

            row = Adw.ActionRow(title=name, subtitle=label)
            row.add_prefix(Gtk.Image.new_from_icon_name(icon))

            is_discovered = contact is not None and not is_known
            if is_discovered:
                row._node_info = {"pub_key": contact.public_key, "type": contact.type, "name": name}
                _add_node_button(row)

            rows_by_prefix[pub_key_hex] = row
            list_box.append(row)

        def on_advert(contact):
            """Also listen for adverts triggered by the discover."""
            # Check if we already have a row for this node from discover response
            existing_row = None
            for prefix, row in rows_by_prefix.items():
                if contact.public_key_hex.startswith(prefix) or prefix.startswith(
                    contact.public_key_hex[:16]
                ):
                    existing_row = row
                    break

            if existing_row:
                # Update existing row with advert data (name, distance)
                adv_name = contact.name or existing_row.get_title()
                existing_row.set_title(adv_name)
                dist = _distance_str(contact)
                if dist:
                    cur_sub = existing_row.get_subtitle() or ""
                    if "km" not in cur_sub and " m" not in cur_sub:
                        existing_row.set_subtitle(f"{cur_sub} · {dist}")
                is_known = self._window.is_contact_known(contact.public_key_hex)
                if hasattr(existing_row, "_node_info"):
                    existing_row._node_info["name"] = adv_name
                    existing_row._node_info["pub_key"] = contact.public_key
                    existing_row._node_info["type"] = contact.type
                elif not is_known:
                    existing_row._node_info = {
                        "pub_key": contact.public_key,
                        "type": contact.type,
                        "name": adv_name,
                    }
                    _add_node_button(existing_row)
                seen.add(contact.public_key_hex)
                return

            if contact.public_key_hex in seen:
                return
            seen.add(contact.public_key_hex)

            icon = type_icons.get(contact.type, "avatar-default-symbolic")
            is_known = self._window.is_contact_known(contact.public_key_hex)
            log.debug(
                "Discover advert: %s has_loc=%s", contact.public_key_hex[:16], contact.has_location
            )

            subtitle = contact.type_label
            dist = _distance_str(contact)
            if dist:
                subtitle += f" · {dist}"
            row = Adw.ActionRow(
                title=contact.name or _("Unknown"),
                subtitle=subtitle,
            )
            row.add_prefix(Gtk.Image.new_from_icon_name(icon))
            if not is_known:
                row._node_info = {
                    "pub_key": contact.public_key,
                    "type": contact.type,
                    "name": contact.name or _("Unknown"),
                }
                _add_node_button(row)
            rows_by_prefix[contact.public_key_hex] = row
            list_box.append(row)

        def on_tick():
            seconds_left[0] -= 1
            if seconds_left[0] <= 0:
                stop_scanning()
                return False
            status_label.set_label(_("Scanning... {}s remaining").format(seconds_left[0]))
            return True

        def stop_scanning():
            self._window.stop_discover()
            self._window.set_advert_listener(None)
            spinner.set_spinning(False)
            count = len(seen)
            status_label.set_label(
                _("Done. {} nodes found.").format(count)
                if count != 1
                else _("Done. {} node found.").format(count)
            )
            if timer_id[0]:
                GLib.source_remove(timer_id[0])
                timer_id[0] = None

        def on_dialog_closed(d):
            stop_scanning()

        dialog.connect("closed", on_dialog_closed)

        # Start scanning - send both discover req and zero-hop advert
        self._window.set_advert_listener(on_advert)
        self._window.send_discover_req(on_discover_response)
        # Also send a zero-hop advert - this is the proven method
        # that triggers nearby nodes to respond with their adverts
        GLib.timeout_add(200, lambda: self._window.send_advert(flood=False, silent=True) or False)
        timer_id[0] = GLib.timeout_add_seconds(1, on_tick)

        dialog.present(self._window)

    def _on_rx_log(self, *args):
        """Open the Rx Log dialog showing received radio packets."""
        builder = Gtk.Builder.new_from_resource("/page/codeberg/sesivany/Meshy/ui/rx-log-dialog.ui")
        dialog = builder.get_object("rx_log_dialog")
        list_box = builder.get_object("rx_log_list")
        clear_btn = builder.get_object("clear_btn")
        placeholder = builder.get_object("rx_log_placeholder")
        list_box.set_placeholder(placeholder)

        sf = self._window.device_info.radio_sf or 12

        def _make_row(entry):
            box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=2)
            box.set_margin_start(12)
            box.set_margin_end(12)
            box.set_margin_top(8)
            box.set_margin_bottom(8)

            header_box = Gtk.Box(spacing=8)
            title = Gtk.Label(
                label=f"{entry.route_type_name} {entry.payload_type_name}", xalign=0, hexpand=True
            )
            title.add_css_class("heading")
            header_box.append(title)
            snr_icon = Gtk.Image.new_from_icon_name(snr_to_icon(entry.snr, sf))
            header_box.append(snr_icon)
            snr_label = Gtk.Label(label=f"{entry.snr:.2f}dB", xalign=1)
            snr_label.add_css_class("dim-label")
            header_box.append(snr_label)
            box.append(header_box)

            def _detail(text, wrap=False):
                lbl = Gtk.Label(label=text, xalign=0)
                lbl.add_css_class("caption")
                lbl.add_css_class("dim-label")
                if wrap:
                    lbl.set_wrap(True)
                    lbl.set_wrap_mode(Pango.WrapMode.WORD_CHAR)
                box.append(lbl)

            _detail(
                f'{entry.timestamp.strftime("%H:%M:%S")} \u00b7 '
                f'{_("Size")}: {entry.size} {_("bytes")}'
            )
            if entry.pkt_hash:
                _detail(f"Hash: {entry.pkt_hash.upper()}")
            path_csv = ",".join(entry.path) if entry.path else ""
            _detail(f'{_("Path")}: {len(entry.path)} {_("hops")} ' f'[{path_csv}]', wrap=True)
            _detail(f'{_("Path Hashes")}: {entry.path_hash_size}-{_("byte per hop")}')
            if entry.transport_from and entry.transport_to:
                _detail(f"From: <{entry.transport_from}> " f"To: <{entry.transport_to}>")

            return box

        for entry in reversed(self._window.rx_log):
            list_box.append(_make_row(entry))

        def on_new_entry(entry):
            list_box.prepend(_make_row(entry))

        self._window.set_rx_log_callback(on_new_entry)

        def on_clear(*_):
            self._window.rx_log.clear()
            child = list_box.get_first_child()
            while child:
                next_c = child.get_next_sibling()
                list_box.remove(child)
                child = next_c

        clear_btn.connect("clicked", on_clear)

        def on_closed(d):
            self._window.set_rx_log_callback(None)

        dialog.connect("closed", on_closed)
        dialog.present(self._window)

    # ─── Firmware update check ──────────────────────────────

    @staticmethod
    def _parse_version(ver_str):
        ver = ver_str.strip().lstrip("v")
        ver = ver.split("-")[0]
        try:
            parts = [int(x) for x in ver.split(".")]
            while len(parts) < 3:
                parts.append(0)
            return tuple(parts)
        except (ValueError, AttributeError):
            return (0, 0, 0)

    def _check_firmware_update(self, current_ver):
        if self._latest_fw_version is not None:
            self._show_update_if_needed(current_ver, self._latest_fw_version)
            return

        def fetch():
            try:
                req = urllib.request.Request(_RELEASES_URL, headers={"User-Agent": "Meshy"})
                with urllib.request.urlopen(req, timeout=10) as resp:
                    releases = json.loads(resp.read())
                for entry in releases:
                    if entry.get("type") == "companion":
                        latest = entry["version"]
                        GLib.idle_add(self._on_releases_fetched, current_ver, latest)
                        return
            except Exception as e:
                log.debug("Firmware update check failed: %s", e)

        threading.Thread(target=fetch, daemon=True).start()

    def _on_releases_fetched(self, current_ver, latest_ver):
        self._latest_fw_version = latest_ver
        self._show_update_if_needed(current_ver, latest_ver)
        return False

    def _show_update_if_needed(self, current_ver, latest_ver):
        current = self._parse_version(current_ver)
        latest = self._parse_version(latest_ver)
        if latest > current:
            self._fw_update_btn.set_visible(True)
            self._fw_update_btn.set_tooltip_text(
                _("Update available: {version}").format(version=latest_ver)
            )
        else:
            self._fw_update_btn.set_visible(False)

    def _on_fw_update_clicked(self, btn):
        launcher = Gtk.UriLauncher.new(_FLASHER_URL)
        launcher.launch(self._window, None, None)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Protocol frame handling extracted from MeshyWindow."""

import logging
import struct
import time
from datetime import datetime

from gi.repository import Gio, GLib
from meshy.models import (
    Channel,
    ChannelMessage,
    Contact,
    ContactType,
    Message,
    MessageStatus,
    NotificationLevel,
)
from meshy.protocol import (
    PUSH_CODE_ADVERT,
    PUSH_CODE_LOG_DATA,
    PUSH_CODE_LOGIN_FAILED,
    PUSH_CODE_LOGIN_SUCCESS,
    PUSH_CODE_MSG_WAITING,
    PUSH_CODE_NEW_ADVERT,
    PUSH_CODE_PATH_DISCOVERY_RESPONSE,
    PUSH_CODE_PATH_UPDATED,
    PUSH_CODE_SEND_CONFIRMED,
    PUSH_CODE_STATUS_RESPONSE,
    PUSH_CODE_TELEMETRY_RESPONSE,
    PUSH_CODE_TRACE_DATA,
    RESP_CODE_ADVERT_PATH,
    RESP_CODE_AUTO_ADD_CONFIG,
    RESP_CODE_BATT_AND_STORAGE,
    RESP_CODE_BINARY_RESPONSE,
    RESP_CODE_CHANNEL_DATA_RECV,
    RESP_CODE_CHANNEL_INFO,
    RESP_CODE_CHANNEL_MSG_RECV,
    RESP_CODE_CHANNEL_MSG_RECV_V3,
    RESP_CODE_CONTACT,
    RESP_CODE_CONTACT_MSG_RECV,
    RESP_CODE_CONTACT_MSG_RECV_V3,
    RESP_CODE_CONTACTS_START,
    RESP_CODE_CONTROL_DATA,
    RESP_CODE_CURR_TIME,
    RESP_CODE_CUSTOM_VARS,
    RESP_CODE_DEFAULT_FLOOD_SCOPE,
    RESP_CODE_DEVICE_INFO,
    RESP_CODE_END_OF_CONTACTS,
    RESP_CODE_ERR,
    RESP_CODE_EXPORT_CONTACT,
    RESP_CODE_NO_MORE_MESSAGES,
    RESP_CODE_OK,
    RESP_CODE_SELF_INFO,
    RESP_CODE_SENT,
    RESP_CODE_STATS,
    STATS_TYPE_RADIO,
    build_get_contacts,
    build_set_device_time,
    identify_response,
    parse_advert_path_response,
    parse_auto_add_config,
    parse_batt_and_storage,
    parse_channel_data,
    parse_channel_info,
    parse_channel_message,
    parse_contact_frame,
    parse_contact_message,
    parse_curr_time,
    parse_custom_vars,
    parse_default_flood_scope,
    parse_device_info,
    parse_discover_response,
    parse_error_response,
    parse_neighbors_response,
    parse_push_send_confirmed,
    parse_self_info,
    parse_sent_response,
    parse_stats_response,
    parse_status_response,
    parse_trace_response,
)

log = logging.getLogger(__name__)


class FrameHandler:
    """Handles incoming protocol frames from the companion device."""

    def __init__(self, win):
        self._win = win
        self._contacts_refresh_id = None
        self._contacts_refetch_id = None
        self._pending_channel_snr = {}
        self._ok_callbacks = []

    def cleanup_on_disconnect(self):
        self._pending_channel_snr.clear()
        for _on_ok, _on_error, timeout_id in self._ok_callbacks:
            GLib.source_remove(timeout_id)
        self._ok_callbacks.clear()

    def handle_frame(self, data: bytes):
        """Process a received protocol frame on the main thread."""
        code = identify_response(data)
        if code is None:
            return False

        # Log all responses when login callback is active
        if self._win._login_callback:
            log.info(
                f"RX during login wait: code=0x{code:02x} ({code}) len={len(data)} hex={data.hex()[:60]}..."
            )

        log.debug(f"RX code={code} len={len(data)} hex={data.hex()}")

        if code == RESP_CODE_OK:
            if self._ok_callbacks:
                on_ok, _on_error, timeout_id = self._ok_callbacks.pop(0)
                GLib.source_remove(timeout_id)
                on_ok()
                return False
            self._win._ignore_next_err = False
            return False

        if code == RESP_CODE_ERR:
            err = parse_error_response(data)
            if self._ok_callbacks:
                _on_ok, on_error, timeout_id = self._ok_callbacks.pop(0)
                GLib.source_remove(timeout_id)
                on_error(err)
                return False
            if self._win._ignore_next_err:
                self._win._ignore_next_err = False
                if err:
                    log.debug(
                        f'Suppressed device error: {err["message"]} (code=0x{err["code"]:02x})'
                    )
                return False
            if err:
                log.warning(f'Device error: {err["message"]} (code=0x{err["code"]:02x})')
                if not self._win._region_callbacks:
                    self._win.show_toast(_("Device error: {}").format(err["message"]))
            if self._win._msg_ctrl.sent_msg_queue:
                front = self._win._msg_ctrl.sent_msg_queue[0]
                if front is None or (isinstance(front, dict) and "type" in front):
                    self._win._msg_ctrl.sent_msg_queue.pop(0)
            return False

        if code == RESP_CODE_SELF_INFO:
            info = parse_self_info(data)
            if self._win._conn_ctrl.gps_poll_pending:
                self._win._conn_ctrl.gps_poll_pending = False
                if info:
                    lat = info.get("adv_lat", 0.0)
                    lon = info.get("adv_lon", 0.0)
                    if abs(lat) > 1e-6 or abs(lon) > 1e-6:
                        self._win._device_info.latitude = lat
                        self._win._device_info.longitude = lon
                    self._win._device_view.update_location(lat, lon)
                    if self._win._map_view_inst is not None:
                        self._win._map_view.update_contacts(self._win._contacts)
                return False
            self._win._conn_ctrl.on_handshake_received()
            if info:
                self._win._self_info = info
                self._win._device_info.public_key = info.get("public_key", "")
                self._win._device_info.name = info.get("name", "")
                self._win._device_info.tx_power = info.get("tx_power", 0)
                self._win._device_info.max_tx_power = info.get("max_tx_power", 0)
                self._win._device_info.radio_freq = info.get("radio_freq", 0)
                self._win._device_info.radio_bw = info.get("radio_bw", 0)
                self._win._device_info.radio_sf = info.get("radio_sf", 0)
                raw_cr = info.get("radio_cr", 0)
                if 1 <= raw_cr <= 4:
                    self._win._device_info.cr_old_encoding = True
                    self._win._device_info.radio_cr = raw_cr + 4
                else:
                    self._win._device_info.cr_old_encoding = False
                    self._win._device_info.radio_cr = raw_cr
                lat = info.get("adv_lat", 0.0)
                lon = info.get("adv_lon", 0.0)
                if abs(lat) > 1e-6 or abs(lon) > 1e-6:
                    self._win._device_info.latitude = lat
                    self._win._device_info.longitude = lon
                self._win._device_view.update_device_info(self._win._device_info)
                self._win._device_view.update_location(
                    info.get("adv_lat", 0.0), info.get("adv_lon", 0.0)
                )
                if self._win._map_view_inst is not None:
                    self._win._map_view.update_contacts(self._win._contacts)
                self._win._settings_view.update_device_info(self._win._device_info)
                self._win._settings_view.update_advert_location(
                    info.get("adv_lat", 0.0),
                    info.get("adv_lon", 0.0),
                    info.get("adv_loc_policy", 0),
                )
                self._win._settings_view.update_telemetry_modes(
                    info.get("telemetry_mode_base", 0),
                    info.get("telemetry_mode_loc", 0),
                    info.get("telemetry_mode_env", 0),
                )
                self._win._device_info.multi_acks = info.get("multi_acks", 0)
                self._win._settings_view.update_multi_acks(info.get("multi_acks", 0))
                self._win._update_companion_label()
                self._win._save_tcp_companion_if_needed()

        elif code == RESP_CODE_CUSTOM_VARS:
            self._win._custom_vars = parse_custom_vars(data)
            self._win._settings_view.update_gps_settings(self._win._custom_vars)
            self._win._device_view.update_gps_enabled(self._win._custom_vars.get("gps") == "1")

        elif code == RESP_CODE_DEVICE_INFO:
            info = parse_device_info(data)
            if info:
                for key in (
                    "model",
                    "ver",
                    "fw_build",
                    "fw_ver",
                    "max_contacts",
                    "max_channels",
                    "path_hash_mode",
                    "repeat",
                    "ble_pin",
                ):
                    if key in info:
                        setattr(self._win._device_info, key, info[key])
                self._win._device_view.update_device_info(self._win._device_info)
                self._win._settings_view.update_device_info(self._win._device_info)

        elif code == RESP_CODE_BATT_AND_STORAGE:
            info = parse_batt_and_storage(data)
            if info:
                self._win._device_info.battery_mv = info.get("level", 0)
                self._win._device_info.storage_used_kb = info.get("used_kb", 0)
                self._win._device_info.storage_total_kb = info.get("total_kb", 0)
                self._win._device_view.update_device_info(self._win._device_info)
                self._win._update_companion_label()

        elif code == RESP_CODE_STATS:
            stats = parse_stats_response(data)
            if stats:
                sub = stats["sub_type"]
                self._win._device_stats[sub] = stats
                self._win._device_view.update_stats(self._win._device_stats)
                if sub == STATS_TYPE_RADIO:
                    total = stats.get("tx_air_secs", 0) + stats.get("rx_air_secs", 0)
                    if total > self._win._msg_ctrl.prev_total_air_secs:
                        self._win._msg_ctrl.last_airtime_bump = time.monotonic()
                    self._win._msg_ctrl.prev_total_air_secs = total

        elif code == RESP_CODE_CONTACTS_START:
            self._win._conn_ctrl.contacts_iter_busy = True
            self._win._storage._batch_depth += 1
            self._win._conn_ctrl._start_contacts_sync_timeout()
            self._win._conn_ctrl.sync_contact_total = (
                int.from_bytes(data[1:3], "little") if len(data) >= 3 else 0
            )
            self._win._conn_ctrl.sync_contact_current = 0
            self._win._conn_ctrl.synced_contact_keys = set()
            if self._win._conn_ctrl.syncing:
                self._win._connection_banner.set_title(
                    _("Syncing contacts") + f"\n{0}/{self._win._conn_ctrl.sync_contact_total}"
                )

        elif code == RESP_CODE_CONTACT:
            self._win._conn_ctrl._start_contacts_sync_timeout()
            contact = parse_contact_frame(data)
            if contact:
                self.add_or_update_contact(contact)
                c = self.dict_to_contact(contact)
                if c:
                    self._win._conn_ctrl.synced_contact_keys.add(c.public_key_hex)
            else:
                log.warning(f"Failed to parse contact frame (raw {len(data)} bytes), skipping")
            self._win._conn_ctrl.sync_contact_current += 1
            if self._win._conn_ctrl.syncing:
                self._win._connection_banner.set_title(
                    _("Syncing contacts")
                    + f"\n{self._win._conn_ctrl.sync_contact_current}/{self._win._conn_ctrl.sync_contact_total}"
                )

        elif code == RESP_CODE_END_OF_CONTACTS:
            self._win._conn_ctrl.contacts_iter_busy = False
            self._win._conn_ctrl._cancel_contacts_sync_timeout()
            if self._win._storage._batch_depth > 0:
                self._win._storage._batch_depth -= 1
                if self._win._storage._batch_depth == 0:
                    self._win._storage._db.commit()
            # Remove contacts no longer on the device (full sync only)
            if (
                self._win._conn_ctrl.synced_contact_keys
                and not self._win._conn_ctrl.incremental_sync
            ):
                synced = self._win._conn_ctrl.synced_contact_keys
                stale_keys = set()
                for c in self._win._contacts:
                    if c.public_key_hex not in synced:
                        stale_keys.add(c.public_key_hex)
                        self._win._contacts_by_key.pop(c.public_key_hex, None)
                        self._win._storage.deactivate_contact(c.public_key_hex)
                self._win._contacts = [
                    c for c in self._win._contacts if c.public_key_hex not in stale_keys
                ]
                self._win._conn_ctrl.synced_contact_keys = None
            self._win._conn_ctrl.incremental_sync = False
            self._win._contacts_view.update_contacts(self._win._contacts)
            if self._win._map_view_inst is not None:
                self._win._map_view.update_contacts(self._win._contacts)
            self._win._refresh_chat_header()
            # Trigger deferred refetch if one was requested during iteration
            if self._win._conn_ctrl.contacts_refetch_pending:
                self._win._conn_ctrl.contacts_refetch_pending = False
                self.schedule_contacts_refetch()
            # Now safe to fetch channels sequentially
            self._win._fetch_channels_sequential(0)

        elif code in (RESP_CODE_CONTACT_MSG_RECV, RESP_CODE_CONTACT_MSG_RECV_V3):
            self._win._msg_ctrl.last_rx_time = time.monotonic()
            msg_data = parse_contact_message(data)
            if msg_data:
                self.handle_incoming_message(msg_data)
                self._win._conn_ctrl.msg_sync_retries = 0
                self._win._send_sync_next_message()

        elif code in (RESP_CODE_CHANNEL_MSG_RECV, RESP_CODE_CHANNEL_MSG_RECV_V3):
            self._win._msg_ctrl.last_rx_time = time.monotonic()
            msg_data = parse_channel_message(data)
            if msg_data:
                self.handle_incoming_channel_message(msg_data)
                self._win._conn_ctrl.msg_sync_retries = 0
                self._win._send_sync_next_message()

        elif code == RESP_CODE_CHANNEL_DATA_RECV:
            result = parse_channel_data(data)
            if result:
                log.debug(
                    f"Channel data received: ch={result['channel_idx']} "
                    f"type={result['data_type']} len={result['data_len']}"
                )

        elif code == RESP_CODE_NO_MORE_MESSAGES:
            self._win._finish_msg_sync()

        elif code == RESP_CODE_SENT:
            result = parse_sent_response(data)
            if result and self._win._msg_ctrl.sent_msg_queue:
                entry = self._win._msg_ctrl.sent_msg_queue.pop(0)
                if isinstance(entry, dict) and "type" in entry:
                    if "expected_ack" in result:
                        tag_hex = result["expected_ack"].hex()
                        self._win._pending_binary_requests[tag_hex] = entry
                        req_type = entry["type"]
                        if req_type == "neighbors":
                            self._win._neighbors_callback = None
                        elif req_type == "acl":
                            self._win._acl_callback = None
                        elif req_type == "telemetry":
                            self._win._telemetry_callback = None
                elif entry is None:
                    # Non-message send (channel, CLI, advert, trace, etc.)
                    if self._win._trace_callback:
                        timeout_ms = result.get("suggested_timeout", 15000)
                        if self._win._trace_timeout_callback:
                            self._win._trace_timeout_callback(timeout_ms)
                elif isinstance(entry, str) and entry.startswith("region:"):
                    # Region discovery request — map ack tag to callback
                    callback_key = entry[7:]  # strip 'region:' prefix
                    if "expected_ack" in result and callback_key in self._win._region_callbacks:
                        ack_hex = result["expected_ack"].hex()
                        self._win._region_callbacks[ack_hex] = self._win._region_callbacks.pop(
                            callback_key
                        )
                elif "expected_ack" in result:
                    msg_id = entry
                    ack_key = result["expected_ack"].hex()
                    if msg_id not in self._win._msg_ctrl.retry_context:
                        # Already delivered by an earlier attempt's ACK — don't
                        # overwrite DELIVERED status with SENT from this late
                        # RESP_CODE_SENT response.
                        pass
                    else:
                        if ack_key != "00000000":
                            self._win._msg_ctrl.pending_acks[ack_key] = msg_id
                            self._win._msg_ctrl._acks_by_msg.setdefault(msg_id, set()).add(ack_key)
                            self._win._msg_ctrl.register_ack_key(msg_id, ack_key)
                        is_flood = result.get("type", 0) == 1
                        detail = self._win._msg_ctrl.send_detail(msg_id, is_flood)
                        self._win._storage.update_message_status(msg_id, MessageStatus.SENT)
                        self._win._chat_view.update_message_status(
                            msg_id, MessageStatus.SENT, detail
                        )
                        est_timeout = result.get("suggested_timeout", 30000)
                        timeout_id = GLib.timeout_add(
                            int(est_timeout * 1.5),
                            self._win._msg_ctrl._on_message_timeout,
                            msg_id,
                            ack_key,
                        )
                        self._win._msg_ctrl.pending_timeouts[msg_id] = timeout_id
            elif result and self._win._trace_callback:
                # Trace sent confirmation with empty queue (fallback)
                timeout_ms = result.get("suggested_timeout", 15000)
                if self._win._trace_timeout_callback:
                    self._win._trace_timeout_callback(timeout_ms)

        elif code == PUSH_CODE_SEND_CONFIRMED:
            result = parse_push_send_confirmed(data)
            if result and "code" in result:
                ack_key = result["code"]
                if ack_key in self._win._msg_ctrl.pending_acks:
                    msg_id = self._win._msg_ctrl.pending_acks.pop(ack_key)
                    # Remove remaining ack_keys for this message
                    remaining = self._win._msg_ctrl._acks_by_msg.pop(msg_id, set())
                    remaining.discard(ack_key)
                    for k in remaining:
                        self._win._msg_ctrl.pending_acks.pop(k, None)
                    if msg_id in self._win._msg_ctrl.pending_timeouts:
                        GLib.source_remove(self._win._msg_ctrl.pending_timeouts.pop(msg_id))
                    trip_ms = result.get("trip_time_ms", 0)
                    self._win._msg_ctrl.record_path_result(
                        msg_id, success=True, trip_ms=trip_ms, ack_key=ack_key
                    )
                    ctx = self._win._msg_ctrl.retry_context.get(msg_id, {})
                    strategy = ctx.get("strategy", "traditional")
                    if strategy in ("override_direct", "override_flood"):
                        self._win._msg_ctrl._restore_original_path(ctx)

                    # For flood success, schedule attribution after refetch
                    path_used = ctx.get("path_used", {})
                    if path_used.get("path_length", 0) < 0:
                        contact = ctx.get("contact")
                        if contact:
                            self._win._msg_ctrl.pending_flood_attributions.setdefault(
                                contact.public_key_hex, []
                            ).append((trip_ms, False))

                    self._win._msg_ctrl.retry_context.pop(msg_id, None)
                    detail = (
                        f"{trip_ms / 1000:.1f}s"
                        if trip_ms and trip_ms >= 1000
                        else f"{trip_ms}ms"
                        if trip_ms
                        else ""
                    )
                    self._win._storage.update_message_status(msg_id, MessageStatus.DELIVERED)
                    self._win._chat_view.update_message_status(
                        msg_id, MessageStatus.DELIVERED, detail
                    )
                    # Re-fetch contacts soon to pick up the newly discovered path
                    self.schedule_contacts_refetch()
                elif ack_key in self._win._msg_ctrl.grace_acks:
                    # Late ACK arrived during grace period — upgrade from failed to delivered
                    msg_id = self._win._msg_ctrl.grace_acks.pop(ack_key)
                    # Remove remaining ack_keys for this message
                    remaining = self._win._msg_ctrl._grace_by_msg.pop(msg_id, set())
                    remaining.discard(ack_key)
                    for k in remaining:
                        self._win._msg_ctrl.grace_acks.pop(k, None)
                    if msg_id in self._win._msg_ctrl.grace_timeouts:
                        GLib.source_remove(self._win._msg_ctrl.grace_timeouts.pop(msg_id))
                    trip_ms = result.get("trip_time_ms", 0)
                    detail = (
                        _("{}s (late)").format(f"{trip_ms / 1000:.1f}")
                        if trip_ms and trip_ms >= 1000
                        else _("{}ms (late)").format(trip_ms)
                        if trip_ms
                        else _("late")
                    )
                    log.info(f"Late ACK received for message {msg_id[:8]}")
                    self._win._storage.update_message_status(msg_id, MessageStatus.DELIVERED)
                    self._win._chat_view.update_message_status(
                        msg_id, MessageStatus.DELIVERED, detail
                    )
                    # Record late path result with reduced weight
                    grace_ctx = self._win._msg_ctrl.grace_contexts.pop(msg_id, None)
                    if grace_ctx:
                        contact_key = grace_ctx["contact_key"]
                        path_used = grace_ctx.get("ack_paths", {}).get(ack_key) or grace_ctx.get(
                            "path_used", {}
                        )
                        if path_used.get("path_length", -1) >= 0:
                            self._win._msg_ctrl.path_history.record_result(
                                contact_key,
                                path_used["path_bytes"],
                                self._win._msg_ctrl._device_hash_size,
                                path_used["hop_count"],
                                success=True,
                                trip_ms=trip_ms,
                                weight_increment=self._win._msg_ctrl.GRACE_WEIGHT_INCREMENT,
                            )
                        else:
                            self._win._msg_ctrl.pending_flood_attributions.setdefault(
                                contact_key, []
                            ).append((trip_ms, True))
                    self.schedule_contacts_refetch()

        elif code == PUSH_CODE_MSG_WAITING:
            self._win._conn_ctrl.msg_sync_retries = 0
            self._win._send_sync_next_message()

        elif code in (PUSH_CODE_ADVERT, PUSH_CODE_NEW_ADVERT):
            self._win._msg_ctrl.last_rx_time = time.monotonic()
            contact_dict = parse_contact_frame(data)
            if contact_dict:
                contact = self.dict_to_contact(contact_dict)
                # Notify live advert listener if active
                if self._win._advert_listener:
                    self._win._advert_listener(contact)
                # If already a known contact, update it
                if self._win.is_contact_known(contact.public_key_hex):
                    self.add_or_update_contact(contact_dict)
                    self._win._contacts_view.update_contacts(self._win._contacts)
                # Always add/update in discovered list (dual-write)
                self.add_or_update_discovered(contact)

        elif code == PUSH_CODE_PATH_UPDATED:
            # Re-fetch contacts to get updated paths (debounced)
            self.schedule_contacts_refetch()

        elif code == RESP_CODE_CHANNEL_INFO:
            ch_info = parse_channel_info(data)
            if ch_info:
                channel = Channel(
                    index=ch_info["channel_idx"],
                    name=ch_info.get("channel_name", ""),
                    psk=ch_info.get("channel_secret", bytes(16)),
                )
                self.add_or_update_channel(channel)
                next_idx = ch_info["channel_idx"] + 1
            else:
                log.warning(f"Failed to parse channel info (raw {len(data)} bytes), skipping")
                next_idx = (
                    self._win._conn_ctrl.channel_sync_index + 1
                    if self._win._conn_ctrl.channel_sync_index >= 0
                    else 0
                )
            self._win._fetch_channels_sequential(next_idx)

        elif code == RESP_CODE_CURR_TIME:
            device_time = parse_curr_time(data)
            if device_time is not None:
                now = int(datetime.now().timestamp())
                drift = now - device_time
                # Auto-sync time on every connection — the official app does this too.
                # A correct RTC is critical for repeater login (anti-replay protection).
                # Firmware rejects setting time backwards, so ERR on device-ahead is
                # expected and harmless — suppress it via _ignore_next_err flag.
                log.info(
                    f"Device clock: {datetime.fromtimestamp(device_time)} "
                    f"(drift {drift}s), sending time sync"
                )
                if drift < 0:
                    self._win._ignore_next_err = True
                self._win.send_frame(build_set_device_time())
                if abs(drift) > 300:
                    mins = abs(drift) // 60
                    if drift > 0:
                        self._win.show_toast(
                            _("Device clock was {}min behind — synced").format(mins)
                        )
                    else:
                        self._win.show_toast(
                            _("Device clock is {}min ahead of system").format(mins)
                        )

        elif code == RESP_CODE_AUTO_ADD_CONFIG:
            config = parse_auto_add_config(data)
            if config:
                self._win._settings_view.update_auto_add_config(config)

        elif code == RESP_CODE_DEFAULT_FLOOD_SCOPE:
            result = parse_default_flood_scope(data)
            if result is not None:
                scope_name = result.get("scope_name", "")
                log.info(f"Default flood scope: " f"{scope_name if scope_name else '(none)'}")

        elif code == RESP_CODE_EXPORT_CONTACT:
            if getattr(self._win, "_pending_qr_callback", None):
                callback = self._win._pending_qr_callback
                self._win._pending_qr_callback = None
                callback(data)
            elif self._win._pending_clipboard_export:
                self._win._pending_clipboard_export = False
                clipboard = self._win.get_clipboard()
                clipboard.set("meshcore://" + data[1:].hex())
                # Persist clipboard content so it survives focus loss (Wayland)
                clipboard.store_async(None, None)

        elif code == RESP_CODE_BINARY_RESPONSE:
            # 0x8C format: [code(1)][reserved(1)][tag(4)][payload...]
            handled = False
            if len(data) >= 6:
                tag_hex = data[2:6].hex()
                region_cb = self._win._region_callbacks.pop(tag_hex, None)
                if region_cb:
                    region_text = data[10:].decode("utf-8", "ignore").strip("\x00")
                    region_cb(region_text)
                    handled = True
                elif tag_hex in self._win._pending_binary_requests:
                    req = self._win._pending_binary_requests.pop(tag_hex)
                    payload = data[6:] if len(data) > 6 else b""
                    if req["type"] == "neighbors":
                        result = parse_neighbors_response(data)
                        req["callback"](result)
                    elif req["type"] == "acl":
                        req["callback"](data)
                    elif req["type"] == "telemetry":
                        req["callback"](payload)
                    handled = True
            if not handled:
                if self._win._neighbors_callback:
                    result = parse_neighbors_response(data)
                    callback = self._win._neighbors_callback
                    self._win._neighbors_callback = None
                    callback(result)
                elif self._win._acl_callback:
                    callback = self._win._acl_callback
                    self._win._acl_callback = None
                    callback(data)
                elif self._win._telemetry_callback:
                    callback = self._win._telemetry_callback
                    self._win._telemetry_callback = None
                    lpp_data = data[6:] if len(data) > 6 else b""
                    callback(lpp_data)

        elif code == PUSH_CODE_TELEMETRY_RESPONSE:
            # 0x8B format: [code(1)][reserved(1)][pubkey_prefix(6)][lpp_data...]
            if self._win._telemetry_callback:
                callback = self._win._telemetry_callback
                self._win._telemetry_callback = None
                lpp_data = data[8:] if len(data) > 8 else b""
                callback(lpp_data)

        elif code == RESP_CODE_CONTROL_DATA:
            result = parse_discover_response(data)
            if result and self._win._discover_callback:
                self._win._discover_callback(result)

        elif code == RESP_CODE_ADVERT_PATH:
            result = parse_advert_path_response(data)
            if result and self._win._advert_path_callback:
                callback = self._win._advert_path_callback
                self._win._advert_path_callback = None
                callback(result)

        elif code == PUSH_CODE_PATH_DISCOVERY_RESPONSE:
            if self._win._path_discovery_callback:
                self._win._path_discovery_callback(data)

        elif code == PUSH_CODE_LOG_DATA:
            self.handle_log_data(data)

        elif code == PUSH_CODE_TRACE_DATA:
            result = parse_trace_response(data)
            if result and self._win._trace_callback:
                self._win._trace_callback(result)

        elif code == PUSH_CODE_LOGIN_SUCCESS:
            log.info(f"Received LOGIN_SUCCESS response, len={len(data)}, hex={data.hex()}")
            # Response format: [code][perms][prefix×6][timestamp×4][...] — 14 bytes
            if len(data) < 8:
                log.warning(f"LOGIN_SUCCESS too short: {len(data)} bytes, expected >= 8")
                return False
            perms = data[1]
            is_admin = (perms & 1) == 1
            log.info(f"Login permissions: 0x{perms:02x}, is_admin={is_admin}")
            response_prefix = data[2:8]
            # Extract repeater clock timestamp if present (bytes 8-11, LE uint32)
            login_timestamp = None
            if len(data) >= 12:
                login_timestamp = struct.unpack_from("<I", data, 8)[0]
                log.info(
                    f"Repeater clock at login: {login_timestamp} "
                    f"({datetime.fromtimestamp(login_timestamp)})"
                )
            if len(data) >= 13:
                acl_permissions = data[12]
                log.info(f"ACL permissions: 0x{acl_permissions:02x}")
            if len(data) >= 14:
                fw_ver_level = data[13]
                log.info(f"Firmware version level: {fw_ver_level}")
            log.info(
                f'Response prefix: {response_prefix.hex()}, expected: {self._win._login_target_prefix.hex() if self._win._login_target_prefix else "none"}'
            )
            if self._win._login_callback:
                if (
                    self._win._login_target_prefix
                    and response_prefix == self._win._login_target_prefix
                ):
                    log.info("Prefix matches! Login successful.")
                    self._win._login_is_admin = is_admin
                    was_flood = self._win._login_was_flood_retry
                    pub_key = self._win._login_pub_key
                    cb = self._win._login_callback
                    self._win.stop_login()
                    if was_flood and pub_key:
                        log.info("Flood login succeeded, clearing stale path for re-discovery")
                        from meshy.protocol import build_reset_path

                        self._win.send_frame(build_reset_path(pub_key))
                    cb(True, login_timestamp, is_admin)
                else:
                    log.warning(
                        f'Prefix mismatch: got {response_prefix.hex()}, expected {self._win._login_target_prefix.hex() if self._win._login_target_prefix else "none"}'
                    )
            else:
                log.warning("LOGIN_SUCCESS received but no callback registered")

        elif code == PUSH_CODE_STATUS_RESPONSE:
            log.info(f"Status response received: {len(data)} bytes")
            result = parse_status_response(data)
            if result and self._win._status_callback:
                log.info(
                    f'Status parsed: battery={result["battery_mv"]}mV uptime={result["uptime_secs"]}s'
                )
                self._win._status_callback(result)
                self._win._status_callback = None
            elif not result:
                log.warning(f"Status parse failed: data too short ({len(data)} bytes)")
            elif not self._win._status_callback:
                log.warning("Status response received but no callback registered")

        elif code == PUSH_CODE_LOGIN_FAILED:
            log.info(f"Received LOGIN_FAILED response, len={len(data)}, hex={data.hex()}")
            # Response format: [code][unknown][6-byte-prefix]
            if len(data) < 8:
                log.warning(f"LOGIN_FAILED too short: {len(data)} bytes, expected >= 8")
                return False
            response_prefix = data[2:8]
            log.info(
                f'Response prefix: {response_prefix.hex()}, expected: {self._win._login_target_prefix.hex() if self._win._login_target_prefix else "none"}'
            )
            if self._win._login_callback:
                if (
                    self._win._login_target_prefix
                    and response_prefix == self._win._login_target_prefix
                ):
                    log.info("Prefix matches! Login failed (wrong password).")
                    cb = self._win._login_callback
                    self._win.stop_login()
                    cb(False, None, False)
                else:
                    log.warning(
                        f'Prefix mismatch: got {response_prefix.hex()}, expected {self._win._login_target_prefix.hex() if self._win._login_target_prefix else "none"}'
                    )
            else:
                log.warning("LOGIN_FAILED received but no callback registered")

        return False  # Don't repeat idle callback

    def expect_ok(self, on_ok, on_error, timeout_ms=5000):
        """Register a callback for the next OK/ERROR response."""
        timeout_id = GLib.timeout_add(timeout_ms, self._on_expect_timeout)
        self._ok_callbacks.append((on_ok, on_error, timeout_id))

    def _on_expect_timeout(self):
        if self._ok_callbacks:
            _on_ok, on_error, _tid = self._ok_callbacks.pop(0)
            on_error({"message": "timeout", "code": 0})
        return False

    def dict_to_contact(self, c: dict) -> Contact | None:
        """Convert meshcore contact dict to Contact model."""
        try:
            path_length = c.get("out_path_len", -1)
            return Contact(
                public_key=bytes.fromhex(c["public_key"]),
                name=c.get("adv_name", "Unknown"),
                type=c.get("type", 1),
                flags=c.get("flags", 0),
                path_length=path_length,
                path=bytes.fromhex(c.get("out_path", "")),
                path_hash_size=c.get("path_hash_size", 1),
                latitude=c.get("adv_lat"),
                longitude=c.get("adv_lon"),
                last_seen=datetime.fromtimestamp(c["last_advert"])
                if c.get("last_advert")
                else None,
                device_lastmod=datetime.fromtimestamp(c["lastmod"]) if c.get("lastmod") else None,
                is_active=True,
            )
        except (ValueError, KeyError) as e:
            log.warning(f"Invalid contact data: {e}")
            return None

    def add_or_update_contact(self, contact_dict: dict):
        """Add or update a contact from a meshcore contact dict."""
        contact = self.dict_to_contact(contact_dict)
        if contact is None:
            return
        key = contact.public_key_hex
        existing = self._win._contacts_by_key.get(key)
        if existing:
            contact.force_flood = existing.force_flood
            contact.auto_reset_path = existing.auto_reset_path
            contact.smart_route_selection = existing.smart_route_selection
            try:
                idx = self._win._contacts.index(existing)
                self._win._contacts[idx] = contact
            except ValueError:
                pass
        else:
            self._win._contacts.append(contact)
        self._win._contacts_by_key[key] = contact
        self._win._storage.save_contact(contact)
        if contact.path_length >= 0:
            pending_list = self._win._msg_ctrl.pending_flood_attributions.pop(key, None)
            if pending_list:
                for pending_trip_ms, late in pending_list:
                    self._win._msg_ctrl.record_flood_attribution(key, pending_trip_ms, late=late)
            else:
                device_hash_size = (
                    getattr(getattr(self._win, "_device_info", None), "path_hash_mode", 0) + 1
                )
                self._win._msg_ctrl.path_history.seed_from_contact(
                    contact.public_key_hex,
                    contact.path.hex(),
                    device_hash_size,
                    contact.path_length,
                )
        self.schedule_contacts_refresh()

    def schedule_contacts_refresh(self):
        """Debounce contact list UI updates — refresh once after a burst of changes."""
        if self._contacts_refresh_id:
            GLib.source_remove(self._contacts_refresh_id)
        self._contacts_refresh_id = GLib.timeout_add(200, self.do_contacts_refresh)

    def do_contacts_refresh(self):
        self._contacts_refresh_id = None
        self._win._contacts_view.update_contacts(self._win._contacts)
        if self._win._map_view_inst is not None:
            self._win._map_view.update_contacts(self._win._contacts)
        self._win._refresh_chat_header()
        return False

    def schedule_contacts_refetch(self):
        """Debounced re-fetch of contacts from device.

        Coalesces multiple rapid requests (e.g. SEND_CONFIRMED + PATH_UPDATED)
        into a single GET_CONTACTS call after a short delay.
        """
        if self._contacts_refetch_id:
            GLib.source_remove(self._contacts_refetch_id)
        self._contacts_refetch_id = GLib.timeout_add(300, self.do_contacts_refetch)

    def do_contacts_refetch(self):
        self._contacts_refetch_id = None
        self.send_get_contacts(incremental=True)
        return False

    def send_get_contacts(self, incremental: bool = False):
        """Send CMD_GET_CONTACTS, deferring if the firmware iterator is busy."""
        if self._win._conn_ctrl.contacts_iter_busy:
            self._win._conn_ctrl.contacts_refetch_pending = True
            return False
        self._win._conn_ctrl.contacts_iter_busy = True
        since = None
        if incremental and self._win._storage:
            since = self._win._storage.get_max_device_lastmod()
        self._win._conn_ctrl.incremental_sync = incremental and since is not None
        self._win.send_frame(build_get_contacts(since=since))
        return False

    def add_or_update_discovered(self, contact: Contact):
        """Add or update a contact in the discovered list (memory + SQLite)."""
        contact.received_at = datetime.now()
        key = contact.public_key_hex
        existing = self._win._discovered_by_key.get(key)
        if existing:
            try:
                idx = self._win._discovered.index(existing)
                self._win._discovered[idx] = contact
            except ValueError:
                pass
        else:
            self._win._discovered.append(contact)
        self._win._discovered_by_key[key] = contact
        if self._win._storage:
            self._win._storage.save_discovered_contact(contact)

    def add_or_update_channel(self, channel: Channel):
        """Add or update a channel, preserving local settings."""
        for i, c in enumerate(self._win._channels):
            if c.index == channel.index:
                # Preserve local settings not stored on device
                channel.unread_count = c.unread_count
                channel.notification_level = c.notification_level
                channel.region = c.region
                self._win._channels[i] = channel
                self._win._storage.save_channel(channel)
                self._win._channels_view.update_channels(self._win._channels)
                self.rebuild_key_store()
                return
        # New channel - check if we have stored settings from a previous session
        stored = self._win._storage.get_channel(channel.index)
        if stored:
            channel.notification_level = stored.notification_level
            channel.region = stored.region
        self._win._channels.append(channel)
        self._win._storage.save_channel(channel)
        self._win._channels_view.update_channels(self._win._channels)
        self.rebuild_key_store()

    def is_contact_chat_visible(self, contact: Contact) -> bool:
        """Check if the chat for this contact is currently visible."""
        if not self._win._chat_view_inst:
            return False
        return (
            self._win._current_view == "contacts"
            and self._win._chat_view_inst._contact is not None
            and self._win._chat_view_inst._contact.public_key_hex == contact.public_key_hex
            and self._win.is_active()
        )

    def is_channel_chat_visible(self, channel_index: int) -> bool:
        """Check if the chat for this channel is currently visible."""
        if not self._win._channels_view_inst:
            return False
        chat = self._win._channels_view_inst.get_chat_widget()
        return (
            self._win._current_view == "channels"
            and chat._current_channel is not None
            and chat._current_channel.index == channel_index
            and self._win.is_active()
        )

    def handle_incoming_message(self, msg_data: dict):
        """Handle an incoming direct message."""
        sender_prefix = msg_data.get("pubkey_prefix", "")
        try:
            contact = self._win._storage.find_contact_by_prefix(bytes.fromhex(sender_prefix))
        except ValueError:
            log.warning(f"Invalid sender prefix: {sender_prefix}")
            return
        if not contact:
            log.warning(f"Message from unknown sender: {sender_prefix}")
            return

        is_cli = msg_data.get("txt_type", 0) == 1
        sender_ts = min(msg_data["sender_timestamp"], time.time())
        text = msg_data["text"]
        room_sender_name = ""

        # Room messages: first 4 bytes of text are the sender's pub key prefix
        if contact.type == ContactType.ROOM and len(text) >= 4 and not is_cli:
            sender_bytes = text[:4].encode("latin-1") if isinstance(text, str) else text[:4]
            text = text[4:]
            # Try to resolve sender name from contacts
            for c in self._win._contacts:
                if c.public_key[:4] == sender_bytes:
                    room_sender_name = c.name
                    break
            if not room_sender_name:
                room_sender_name = sender_bytes.hex()

        if contact.type == ContactType.CHAT and self._win._storage.has_duplicate_message(
            contact.public_key_hex, sender_ts, text
        ):
            log.debug(f"Dropping duplicate message from {contact.name}")
            return

        message = Message(
            sender_key=contact.public_key,
            text=text,
            timestamp=datetime.fromtimestamp(sender_ts),
            is_outgoing=False,
            is_cli=is_cli,
            status=MessageStatus.DELIVERED,
            room_sender_name=room_sender_name,
        )
        # Route prefixed CLI responses to their registered callback
        if (
            is_cli
            and len(text) > 3
            and text[2] == "|"
            and text[:3] in self._win._cli_pending_prefixes
        ):
            prefix = text[:3]
            cb = self._win._cli_pending_prefixes.pop(prefix)
            cb(text[3:])
            return

        # Route CLI responses to management dialog if open
        if is_cli and self._win._cli_response_callback:
            self._win._cli_response_callback(contact, message)
            return

        # Repeaters and sensors don't have a chat view – silently drop
        # any non-CLI messages so no phantom unread badge can appear.
        if contact.type in (ContactType.REPEATER, ContactType.SENSOR):
            log.debug(f"Dropping non-CLI message from {contact.type.name} {contact.name}")
            return

        self._win._storage.save_message(contact.public_key_hex, message)
        contact.last_message_at = message.timestamp
        self.schedule_contacts_refresh()
        self._win._chat_view.on_message_received(contact, message)

        if not self.is_contact_chat_visible(contact):
            # Initialize last_read_at if not yet set so the divider has a reference
            if self._win._storage.get_contact_last_read_at(contact.public_key_hex) is None:
                self._win._storage.set_contact_last_read_at(
                    contact.public_key_hex,
                    message.timestamp.timestamp() - 1,
                )
            self._win._contacts_view.mark_unread(contact.public_key_hex, True)
            if self._win._conn_ctrl.msg_syncing:
                self._win._conn_ctrl.sync_msg_count += 1
            else:
                self._win._send_notification(
                    contact.name,
                    message.text,
                    f"contact-{contact.public_key_hex}",
                )

    def handle_incoming_channel_message(self, msg_data: dict):
        """Handle an incoming channel message."""
        my_name = self._win._self_info.get("name", "")
        sender_name = msg_data.get("sender_name", "")
        msg_text = msg_data.get("msg_text", msg_data.get("text", ""))
        raw_sender_ts = msg_data["sender_timestamp"]
        sender_ts = min(raw_sender_ts, time.time())
        channel_idx = msg_data["channel_idx"]

        # Detect echo/repeat of our own outgoing message
        if my_name and sender_name == my_name:
            snr = msg_data.get("SNR")
            path_len = msg_data.get("path_len", -1)
            log.info(
                f"Echo detected: ch={channel_idx} ts={sender_ts} "
                f'text="{msg_text[:30]}" snr={snr} path_len={path_len}'
            )
            path_hops = []
            path_bytes_hex = msg_data.get("path_bytes", "")
            path_len_val = msg_data.get("path_len", 0)
            if path_bytes_hex and path_len_val > 0:
                hash_size = len(path_bytes_hex) // (2 * path_len_val) if path_len_val else 1
                chars = hash_size * 2
                path_hops = [
                    path_bytes_hex[i : i + chars].upper()
                    for i in range(0, len(path_bytes_hex), chars)
                ]
            self._win._channels_view.on_repeat_received(
                channel_idx,
                sender_ts,
                msg_text,
                path_hops,
            )
            return

        channel_msg = ChannelMessage(
            channel_index=channel_idx,
            sender_name=sender_name,
            text=msg_text,
            timestamp=datetime.fromtimestamp(sender_ts),
        )

        # Cache SNR for LOG_DATA to pick up (use raw timestamp to match
        # the key used by LOG_DATA before clamping)
        snr = msg_data.get("SNR")
        if snr is not None:
            cache_key = (channel_idx, raw_sender_ts, msg_text)
            self._pending_channel_snr[cache_key] = snr

        # Check if this is a duplicate (same message via different repeater)
        is_dup = self._win._channels_view.on_path_variant_received(
            channel_msg,
        )
        if is_dup:
            return  # Don't save or notify for duplicate copies

        self._win._storage.save_channel_message(channel_msg)
        self._win._channels_view.on_channel_message_received(channel_msg)

        # Find the channel for notification settings
        channel = None
        for ch in self._win._channels:
            if ch.index == channel_msg.channel_index:
                channel = ch
                break

        if not self.is_channel_chat_visible(channel_msg.channel_index):
            # Initialize last_read_at if not yet set so the divider has a reference
            if self._win._storage.get_channel_last_read_at(channel_msg.channel_index) is None:
                self._win._storage.set_channel_last_read_at(
                    channel_msg.channel_index,
                    channel_msg.timestamp.timestamp() - 1,
                )
            self._win._channels_view.mark_unread(channel_msg.channel_index, True)

            # Send notification based on channel settings
            if channel:
                level = channel.notification_level
                if level == NotificationLevel.DEFAULT:
                    settings = Gio.Settings(schema_id="page.codeberg.sesivany.Meshy")
                    level = NotificationLevel(
                        settings.get_int("default-channel-notification-level")
                    )
                should_notify = False
                if level == NotificationLevel.ALL:
                    should_notify = True
                elif level == NotificationLevel.MENTIONS:
                    if my_name and my_name.lower() in channel_msg.text.lower():
                        should_notify = True

                if should_notify:
                    if self._win._conn_ctrl.msg_syncing:
                        self._win._conn_ctrl.sync_msg_count += 1
                    else:
                        ch_name = channel.name or _("Channel {}").format(channel.index)
                        sender = channel_msg.sender_name or _("Someone")
                        self._win._send_notification(
                            ch_name,
                            _("{sender}: {text}").format(sender=sender, text=channel_msg.text),
                            f"channel-{channel_msg.channel_index}",
                        )

    def restore_unread_counts(self):
        """Restore unread message counts from storage on startup."""
        if not self._win._storage:
            return
        for contact in self._win._contacts:
            # Repeaters and sensors have no chat view – skip them.
            if contact.type in (ContactType.REPEATER, ContactType.SENSOR):
                continue
            count = self._win._storage.count_unread_messages(contact.public_key_hex)
            if count > 0:
                if self._win._storage.get_contact_last_read_at(contact.public_key_hex) is None:
                    oldest = self._win._storage.get_oldest_unread_contact_timestamp(
                        contact.public_key_hex
                    )
                    ts = (oldest - 1) if oldest is not None else (datetime.now().timestamp() - 1)
                    self._win._storage.set_contact_last_read_at(contact.public_key_hex, ts)
                self._win._contacts_view._unread_counts[contact.public_key_hex] = count
                self._win._contacts_view._unread_keys.add(contact.public_key_hex)
        # Refresh to show badges
        self._win._contacts_view.update_contacts(self._win._contacts)

        for channel in self._win._channels:
            if channel.is_empty:
                continue
            count = self._win._storage.count_unread_channel_messages(channel.index)
            if count > 0:
                # Initialize last_read_at if not set, so the unread divider
                # is placed correctly when the user opens the channel.
                if self._win._storage.get_channel_last_read_at(channel.index) is None:
                    oldest = self._win._storage.get_oldest_unread_channel_timestamp(channel.index)
                    ts = (oldest - 1) if oldest is not None else (datetime.now().timestamp() - 1)
                    self._win._storage.set_channel_last_read_at(channel.index, ts)
                self._win._channels_view._unread_counts[channel.index] = count
                self._win._channels_view._unread_indices.add(channel.index)
        # Refresh to show badges
        self._win._channels_view.update_channels(self._win._channels)

        self._win.update_nav_unread_contacts()
        self._win.update_nav_unread_channels()

    def rebuild_key_store(self):
        """Rebuild the channel hash->PSK map from current channels."""
        from meshy.mesh_crypto import channel_hash

        self._win._channel_hash_map = {}
        self._win._channel_psk_map = {}  # hash -> psk bytes
        for ch in self._win._channels:
            if ch.is_empty:
                continue
            h = channel_hash(ch.psk_hex)
            self._win._channel_hash_map[h] = ch.index
            self._win._channel_psk_map[h] = ch.psk

    def handle_log_data(self, data: bytes):
        """Handle LOG_DATA (0x88) push - raw radio packet with path bytes."""
        from meshy.mesh_crypto import (
            PAYLOAD_ADVERT,
            PAYLOAD_GROUP_TEXT,
            decode_log_packet,
            decrypt_group_text,
        )
        from meshy.models import RxLogEntry

        if len(data) < 4:
            return

        snr = struct.unpack("b", data[1:2])[0] / 4.0
        rssi = struct.unpack("b", data[2:3])[0]
        mesh_data = data[3:]

        try:
            result = decode_log_packet(mesh_data)
        except Exception as e:
            log.warning(f"LOG_DATA decode failed: {e}")
            return

        if result:
            entry = RxLogEntry(
                timestamp=datetime.now(),
                snr=snr,
                rssi=rssi,
                payload_type=result["payload_type"],
                route_type=result.get("route_type", 0),
                path=result.get("path", []),
                path_hash_size=result.get("path_hash_size", 1),
                pkt_hash=result.get("pkt_hash", ""),
                size=len(mesh_data),
                transport_from=result.get("transport_from", ""),
                transport_to=result.get("transport_to", ""),
                raw_hex=mesh_data.hex(),
            )
            self._win._rx_log.append(entry)
            if self._win._rx_log_callback:
                self._win._rx_log_callback(entry)

            if entry.path:
                key = entry.path[-1]
                now = time.monotonic()
                existing = self._win._direct_repeaters.get(key)
                best = max(snr, existing[0]) if existing else snr
                self._win._direct_repeaters[key] = (best, now, 0)
                self._win._repeaters_refreshed.add(key)
                stale = [
                    k for k, (_, ts, _mc) in self._win._direct_repeaters.items() if now - ts > 600
                ]
                for k in stale:
                    del self._win._direct_repeaters[k]
                if len(self._win._direct_repeaters) > 5:
                    weakest = min(
                        self._win._direct_repeaters, key=lambda k: self._win._direct_repeaters[k][0]
                    )
                    del self._win._direct_repeaters[weakest]
                if not self._win._signal_icon_timer_id:
                    self._win._signal_icon_timer_id = GLib.timeout_add_seconds(
                        1, self._win._update_signal_icon_tick
                    )

        if not result:
            return

        if result["payload_type"] == PAYLOAD_ADVERT and "advert" in result:
            advert = result["advert"]
            self_key = self._win._self_info.get("public_key", "")
            if advert["public_key"] != self_key:
                contact = self.dict_to_contact(advert)
                if contact:
                    self.add_or_update_discovered(contact)

        if result["payload_type"] != PAYLOAD_GROUP_TEXT:
            return

        if not self._win._channel_hash_map:
            return

        ch_hash = result.get("channel_hash")
        ciphertext = result.get("ciphertext")
        if not ch_hash or not ciphertext:
            return

        psk = self._win._channel_psk_map.get(ch_hash)
        if not psk:
            return

        decrypted = decrypt_group_text(ciphertext, psk)
        if not decrypted:
            log.warning(
                f"LOG_DATA: decryption failed, "
                f"channel_hash={ch_hash}, "
                f"available_channels={[ch.index for ch in self._win._channels if not ch.is_empty]}"
            )
            return

        sender = decrypted.get("sender", "")
        text = decrypted.get("message", "")
        raw_timestamp = decrypted.get("timestamp", 0)
        timestamp = min(raw_timestamp, time.time())
        path_hops = result.get("path", [])

        if not sender or not text:
            return

        path_bytes_hex = "".join(h.lower() for h in path_hops)
        my_name = self._win._self_info.get("name", "")

        log.info(
            f'LOG_DATA: sender="{sender}" path={path_hops} ' f'text="{text[:30]}..." ts={timestamp}'
        )

        # Match channel using pre-built hash map
        channel_idx = self._win._channel_hash_map.get(ch_hash.lower())
        if channel_idx is None:
            log.debug(f"LOG_DATA: no channel match for hash={ch_hash}")
            return

        # Pull SNR cached by CHANNEL_MSG_RECV handler (use raw timestamp
        # to match the key stored by CHANNEL_MSG_RECV before clamping)
        cache_key = (channel_idx, raw_timestamp, text)
        cached_snr = self._pending_channel_snr.pop(cache_key, None)

        path_info = {
            "snr": cached_snr,
            "path_len": len(path_hops),
            "path_bytes": path_bytes_hex,
        }

        if my_name and sender == my_name:
            # Our own message echoed back through repeaters
            original_key = self._win._channels_view.on_repeat_received(
                channel_idx,
                timestamp,
                text,
                path_hops,
            )
            store_ts = original_key[1] if original_key else timestamp
            if self._win._storage:
                self._win._storage.save_channel_message_path(
                    channel_idx,
                    store_ts,
                    text,
                    path_info,
                )
        else:
            # Someone else's message - record path variant
            channel_msg = ChannelMessage(
                channel_index=channel_idx,
                sender_name=sender,
                text=text,
                timestamp=datetime.fromtimestamp(timestamp),
            )
            if self._win._storage:
                self._win._storage.save_channel_message_path(
                    channel_idx,
                    timestamp,
                    text,
                    path_info,
                )
            self._win._channels_view.add_path_info(
                channel_msg,
                path_info,
            )



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Platform-agnostic BLE transport base class."""

from abc import ABC, abstractmethod
from collections.abc import Callable

from meshy.transport_base import ConnectionState

# Nordic UART Service UUIDs (same across all platforms)
NUS_SERVICE_UUID = "6e400001-b5a3-f393-e0a9-e50e24dcca9e"
NUS_TX_CHAR_UUID = "6e400003-b5a3-f393-e0a9-e50e24dcca9e"  # Notify (device -> app)
NUS_RX_CHAR_UUID = "6e400002-b5a3-f393-e0a9-e50e24dcca9e"  # Write (app -> device)


class BleDevice:
    """Platform-agnostic BLE device representation."""

    def __init__(self, path: str, address: str, name: str, rssi: int = -100):
        self.path = path
        self.address = address
        self.name = name
        self.rssi = rssi

    def __repr__(self):
        return f"BleDevice({self.name}, {self.address})"


class BleTransportBase(ABC):
    """Abstract base class for platform-specific BLE implementations.

    Implementations must support the Nordic UART Service (NUS) for
    communication with MeshCore devices.
    """

    # Callbacks that consumers can set
    on_state_changed: Callable[[ConnectionState], None] | None = None
    on_data_received: Callable[[bytes], None] | None = None
    on_device_discovered: Callable[[BleDevice], None] | None = None
    on_passkey_requested: Callable[[str, Callable], None] | None = None

    @property
    @abstractmethod
    def state(self) -> ConnectionState:
        """Current connection state."""
        ...

    @property
    @abstractmethod
    def is_connected(self) -> bool:
        """True if connected to a device."""
        ...

    @property
    @abstractmethod
    def connected_address(self) -> str | None:
        """Address of the connected device, or None."""
        ...

    @property
    @abstractmethod
    def discovered_devices(self) -> list[BleDevice]:
        """List of discovered BLE devices."""
        ...

    @abstractmethod
    def start_scan(self):
        """Start scanning for BLE devices."""
        ...

    @abstractmethod
    def stop_scan(self):
        """Stop scanning for BLE devices."""
        ...

    @abstractmethod
    def connect_device(self, device_path: str):
        """Connect to a device by its path/identifier."""
        ...

    @abstractmethod
    def connect_by_address(self, address: str):
        """Connect to a device by its MAC address."""
        ...

    @abstractmethod
    def pair_device(self, device_path: str):
        """Pair with a device."""
        ...

    @abstractmethod
    def disconnect(self):
        """Disconnect from the current device."""
        ...

    @abstractmethod
    def send_data(self, data: bytes):
        """Send data to the connected device."""
        ...

    @abstractmethod
    def is_bluetooth_enabled(self) -> bool:
        """Check if Bluetooth is enabled."""
        ...

    @abstractmethod
    def has_bluetooth_adapter(self) -> bool:
        """Check if a Bluetooth adapter is available."""
        ...

    @abstractmethod
    def get_paired_devices(self) -> list[BleDevice]:
        """Get list of paired devices that support NUS."""
        ...

    @abstractmethod
    def remove_device(self, device_path: str) -> bool:
        """Remove/unpair a device."""
        ...
